from inspect import isclass
from furthrmind.collection.baseclass import BaseClass
from typing_extensions import List, Self


class Category(BaseClass):
    id = ""
    name = ""
    description = ""
    project = ""

    _attr_definition = {
        "project": {"class": "Project"}
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    def _get_url_instance(self, project_id=None):
        project_url = Category.fm.get_project_url(project_id)
        url = f"{project_url}/researchcategory/{self.id}"
        return url

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/researchcategory/{id}"
        return url

    @classmethod
    def _get_all_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/researchcategory"
        return url

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/researchcategory"
        return url

    @classmethod
    def get(cls, id: str = "", project_id: str = "") -> Self:
        """
        Parameters
        ----------
        id : str
            The id of the requested category. Only needed if used on as a class method
        project_id : str
            The project_id parameter is optional and can be used to retrieve categories
            from another project as the furthrmind sdk was initiated with.

        Returns
        -------
        Self
            An instance of the category class.

        Raises
        ------
        AssertionError
            If used as a class method and id is not specified.

        Example Usage
        -------------
        # Get a category by id
        category = Category.get(id='category_id')

        # Get a category by id using class method
        category = Category.get('category_id')

        # Get a category using the id of the instance
        category = category.get()
        """

        if isclass(cls):
            assert id, "id must be specified"
        return super().get(id=id, project_id=project_id)


    # noinspection PyMethodOverriding
    @classmethod
    def get_many(cls, ids: List[str] = (), names: List[str] = (), project_id: str = "") -> List[Self]:
        """
        Parameters
        ----------
        ids : List[str]
            List with ids.
        names : List[str]
            List with names.
        project_id : str, optional
            Optionally, to get experiments from another project as the furthrmind sdk was initiated with. Defaults to None.

        Returns
        -------
        List[Self]
            List with instances of the category class.

        Raises
        ------
        AssertionError
            If ids or names are not specified.
        """

        assert ids or names, "ids or names must be specified"
        return cls._get_many(ids, names, project_id=project_id)

    @classmethod
    def get_all(cls, project_id: str = "") -> List[Self]:
        """
        Parameters
        ----------
        project_id : str (optional)
            Optionally to get categories from another project as the furthrmind sdk was initiated with, defaults to None

        Returns
        -------
        List[Self]
            List with instances of category class
        """

        return cls._get_all(project_id=project_id)
    
    @staticmethod
    def create(name: str, project_id: str = "") -> Self:
        """
        Method to create a new category

        Parameters
        ----------
        name : str
            Name of the new category

        project_id : str, optional
            Identifier of another project where the category should be created,
            defaults to an empty string

        Returns
        -------
        Self
            The newly created category object
        """

        pass







import math

import numpy as np
import pandas
from typing_extensions import Self, List, Dict
from inspect import isclass
from ..utils import furthr_wrap
from furthrmind.collection.baseclass import BaseClass
from furthrmind.collection.fielddata import FieldData


class Column(BaseClass):
    id = ""
    name = ""
    type = ""
    values = []


    _attr_definition = {
        "columns": {"class": "Column"}
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    @classmethod
    def get(cls, id: str = "", project_id: str = "") -> Self:

        """
        Method to get one column by its id
        If called on an instance of the class, the id of the instance is used

        Parameters
        ----------
        id : str
            id of requested column
        project_id : str, optional
            Optionally to get a column from another project as the furthrmind sdk was initiated with, defaults to ""

        Returns
        -------
        Self
            Instance of column class

        Raises
        ------
        AssertionError
            If used as a class method and id is not specified.
        """

        if isclass(cls):
            assert id, "id must be specified"

        cls._get(id, project_id=project_id)

    # noinspection PyMethodOverriding
    @classmethod
    def get_many(cls, ids: List[str] = (), project_id: str = "") -> List[Self]:
        """
        Method to get many columns belonging to one project

        Parameters
        ----------
        ids : List[str]
            List with ids.

        project_id : str
            Optionally, the id of the project from which to get the experiments. Defaults to an empty string.

        Returns
        -------
        List[Self]
            List with instances of the experiment class.

        Raises
        ------
        AssertionError
            If `ids` is not specified.
        """

        assert ids, "ids must be specified"
        return cls._get_many(ids, project_id=project_id)

    @classmethod
    def _get_all(cls, project_id=None) -> List[Self]:
        raise ValueError("Not implemented for columns")

    def _get_url_instance(self, project_id=None):
        project_url = Column.fm.get_project_url(project_id)
        url = f"{project_url}/column/{self.id}"
        return url

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/column/{id}"
        return url

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/column"
        return url

    @classmethod
    def _get_all_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/column"
        return url

    @classmethod
    def _type_check(cls, column_type, data):
        if not column_type in ["Text", "Numeric"]:
            raise ValueError("Column type must be Text/Numeric")
        if isinstance(data, pandas.Series):
            data = data.tolist()

        new_data = []
        for d in data:
            try:
                if math.isnan(d):
                    new_data.append(None)
                else:
                    new_data.append(d)
            except:
                new_data.append(d)
            # data = [d if not math.isnan(d) else None for d in data]
        data = new_data
        if column_type == "Text":
            if all([isinstance(d, (str, type(None))) for d in data]):
                return data
            return [str(d) for d in data]

        elif column_type == "Numeric":
            if all([isinstance(d, (int, float, type(None))) for d in data]):
                return data
            new_data = []
            for d in data:
                if d is None:
                    new_data.append(d)
                try:
                    new_data.append(float(d))
                except:
                    raise ValueError("All column values must be a float, int or a string that can be converted to a float")
            return new_data

    @classmethod
    @BaseClass._create_instances_decorator(_fetched=False)
    def create(cls, name: str, type: str, data: list, unit=None, project_id=None) -> Self:
        """
        Method to create a new column

        Parameters
        ----------
        name : str
            Name of the column
        type : str
            Type of the column, Either "Text" or "Numeric". Data must fit to type, for Text all data
            will be converted to string and for Numeric all data is converted to float (if possible)
        data : Union[list, pandas.Series]
            List or pandas series. values, must fit to column_type.
        unit : Optional[Union[str, int, dict]]
            Dict with id or name, or name as string, or id as string
        project_id : Optional[str]
            Optionally to create an item in another project as the furthrmind sdk was initiated with

        Returns
        -------
        Self
            Instance of column class

        """
        data = cls._type_check(type, data)
        unit = FieldData._check_unit(unit)
        data_dict = {"name": name, "type": type, "values": data, "unit": unit}
        id = cls._post(data_dict, project_id)
        data_dict["id"] = id
        return data_dict


    @classmethod
    @BaseClass._create_instances_decorator(_fetched=False)
    def create_many(cls, data_list: List[Dict], project_id: str = "") -> List[Self]:
        """
        Method to create many new columns

        Parameters
        ----------
        data_list : List[Dict]
            A list of dictionaries containing information about the data columns to be created. Each dictionary should have the following keys:
                - name: Name of the column
                - type: Type of the column. Allowed values are "Text" or "Numeric".
                - data: List of column values. The values must match the column type. Can also be a pandas data series.
                - unit: Optional. Dictionary with id or name, or name as a string, or id as a string.

        project_id : str, optional
            Optionally to create an columns in another project as the furthrmind sdk was initiated with

        Returns
        -------
        List[Self]
            A list of instances of the column class.

        """

        new_data_list = []
        for item in data_list:
            type = item.get("type")
            unit = item.get("unit")
            name = item.get("name")
            data = item.get("data")
            data = cls._type_check(type, data)
            unit = FieldData._check_unit(unit)
            data_dict = {"name": name, "type": type, "values": data, "unit": unit}
            new_data_list.append(data_dict)

        id_list = cls._post(new_data_list, project_id, force_list=True)
        for item, id in zip(new_data_list, id_list):
            item["id"] = id

        return new_data_list


from furthrmind.collection.baseclass import BaseClass
from typing_extensions import List, Self, Dict, TYPE_CHECKING
from inspect import isclass

if TYPE_CHECKING:
    from furthrmind.collection.comboboxentry import ComboBoxEntry


class Field(BaseClass):
    id = ""
    name = ""
    type = ""
    script = ""
    comboboxentries: List["ComboBoxEntry"] = []

    _attr_definition = {
        "comboboxentries": {"class": "ComboBoxEntry"}
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    def _get_url_instance(self, project_id=None):
        project_url = Field.fm.get_project_url(project_id)
        url = f"{project_url}/fields/{self.id}"
        return url

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/fields/{id}"
        return url

    @classmethod
    def _get_all_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/fields"
        return url

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/fields"
        return url

    @classmethod
    def get(cls, id: str = "", name: str = "", project_id: str = "") -> Self:
        """
        Method to get all one field by its id or name
        If called on the instance of a class, the id of the instance is used

        Parameters
        ----------
        id : str
            id of the requested field
        name : str
            name of the requested field
        project_id : str, optional
            Optionally to get a field from another project as the furthrmind sdk was initiated with

        Returns
        -------
        Self
            Instance of the field class

        Raises
        ------
        AssertionError
            If called as class method and neither id nor name is specified
        """

        if isclass(cls):
            assert id or name, "Either id or name must be specified"
        return cls._get(id, name=name, project_id=project_id)

    @classmethod
    def get_many(cls, ids: List[str] = (), names: List[str] = (), project_id: str = "") -> List[Self]:
        """
        Method to get many fields belonging to one project

        Parameters
        ----------
        ids : list of str, optional
            List of ids to get fields for.

        names : list of str, optional
            List of names to get fields for.

        project_id : str, optional
            Optionally to get fields from another project as the furthrmind sdk was initiated with

        Returns
        -------
        list[Self]
            List of instances of the experiment class.

        Raises
        ------
        AssertionError
            If neither ids nor names are specified.

        """

        assert ids or names, "Either ids or names must be specified"
        return cls._get_many(ids, names=names, project_id=project_id)

    @classmethod
    def get_all(cls, project_id: str = "") -> List[Self]:
        """
        Method to get all fields belonging to one project

        Parameters
        ----------
        project_id : str, optional
            Optionally to get fields from another project as the furthrmind sdk was initiated with

        Returns
        -------
        List[Self]
            A list of instances of the field class, representing all the fields belonging to the specified project.
        """

        return cls._get_all(project_id=project_id)

    @classmethod
    @BaseClass._create_instances_decorator(_fetched=False)
    def create(cls, name: str, type: str, project_id: str = "") -> Self:

        """
        Method to create a new field

        Parameters
        ----------
        name : str
            The name of the field to be created.
        type : str
            The field type of the field. Must be one of the following. Lower or uppercase is not considered,
                i.e. Numeric or numeric is both valid:
                - numeric, numeric-field, numeric_field
                - date, date_field, date-field, datefield
                - singleline, singlelinefield, text, text-field, text_field, textfield
                - combobox, comboboxfield, list, list-field, list_field, listfield
                - multiline, notebook, notebookfield, notebook-field, notebook_field
                - checkbox, checkbox-field, checkbox_field, checkboxfield
                - calculation, calculation-field, calculation_field, calculationfield
        project_id : str, optional
            Optionally, the ID of the project to create an item in. If not provided, the item will be created in the project associated with the furthrmind SDK.

        Returns
        -------
        instance of the sample class
            The created sample instance.

        Raises
        ------
        ValueError
            If name is empty or if type is not one of the allowed types.

        """

        if not name:
            raise ValueError("Name cannot be empty")

        type = cls._check_field_type(type)
        data = {"name": name, "type": type}
        id = cls._post(data, project_id)
        data["id"] = id
        return data

    @classmethod
    def _check_field_type(cls, field_type: str):
        field_type_mapping = {
            "Numeric": ["numeric", "numeric-field", "numeric_field", "numericfield"],
            "Date": ["date", "date_field", "date-field", "datefield"],
            "SingleLine": ["singleline", "single-line", "single_line", "singlelinefield", "single-line-field",
                           "single_line_field", "text", "text-field", "text_field", "textfield"],
            "ComboBox": ["combobox", "comboboxfield", "combobox-field", "combobox_field", "comboboxentry",
                         "list", "listfield", "list-field", "list_field"],
            "MultiLine": ["multiline", "multi_line", "mulit-line", "multilinefield", "multi-line-field",
                          "multi_line_field", "notebook-field", "notebook_field", "notebookfield"],
            "CheckBox": ["checkbox", "checkbox-field", "checkbox_field", "checkboxfield"],
            "Calculation": ["calculation", "calculation-field", "calculation_field", "calculationfield"],
        }
        for field_type_server in field_type_mapping:
            if field_type.lower() in field_type_mapping[field_type_server]:
                return field_type_server

        raise ValueError("Wrong field type")

    @classmethod
    @BaseClass._create_instances_decorator(_fetched=False)
    def create_many(cls, data_list: List[Dict], project_id=None) -> List[Self]:
        """
        Method to create multiple fields

        Parameters
        ----------
        data_list : List[Dict]
            A list of dictionaries containing the data for creating multiple samples. Each dictionary should
            have the following keys:
            - name : str
                The name of the field to be created.
            - type : str
                The field type of the field. Must be one of the following. Lower or uppercase is not considered,
                i.e. Numeric or numeric is both valid:
                - numeric, numeric-field, numeric_field
                - date, date_field, date-field, datefield
                - singleline, singlelinefield, text, text-field, text_field, textfield
                - combobox, comboboxfield, list, list-field, list_field, listfield
                - multiline, notebook, notebookfield, notebook-field, notebook_field
                - checkbox, checkbox-field, checkbox_field, checkboxfield
                - calculation, calculation-field, calculation_field, calculationfield
        project_id : int, optional
            Optionally to create fields in another project as the furthrmind sdk was initiated with

        Returns
        -------
        List[Self]
            A list of instance of the field class that have been created.

        Raises
        ------
        ValueError
            When the "name" key is missing or empty in any of the dictionaries in data_list.
        ValueError
            When the "type" value is not one of: Numeric, Date, SingleLine, ComboBox, MultiLine, CheckBox, Calculation.


        Examples
        --------
        data_list = [
            {"name": "Field 1", "type": "Numeric"},
            {"name": "Field 2", "type": "SingleLine"},
            {"name": "Field 3", "type": "Date"},
            {"name": "Field 4", "type": "ComboBox"},
        ]

        fields = ClassName.create_many(data_list)

        """

        for data in data_list:
            if not "name" in data:
                raise ValueError("Name cannot be empty")

            field_type = cls._check_field_type(data.get("type"))
            data["type"] = field_type

        id_list = cls._post(data_list, project_id, force_list=True)
        for data, id in zip(data_list, id_list):
            data["id"] = id
        return data_list



from inspect import isclass

from typing_extensions import Self, Dict, List, TYPE_CHECKING

from furthrmind.collection.baseclass import (BaseClassWithFieldData, BaseClassWithFiles,
                                             BaseClassWithGroup, BaseClass,
                                             BaseClassWithLinking)
from furthrmind.utils import instance_overload

if TYPE_CHECKING:
    from furthrmind.collection import *


class Sample(BaseClassWithFieldData,
             BaseClassWithFiles, BaseClassWithGroup,
             BaseClassWithLinking, BaseClass):
    id = ""
    name = ""
    neglect = False
    shortid = ""
    files: List["File"] = []
    fielddata: List["FieldData"] = []
    linked_experiments: List["Experiment"] = []
    linked_samples: List[Self] = []
    linked_researchitems: Dict[str, List["ResearchItem"]] = {}
    groups: List["Group"] = []
    datatables: List["DataTable"] = []

    _attr_definition = {
        "files": {"class": "File"},
        "fielddata": {"class": "FieldData"},
        "groups": {"class": "Group"},
        "linked_samples": {"class": "Sample"},
        "linked_experiments": {"class": "Experiment"},
        "linked_researchitems": {"class": "ResearchItem", "nested_dict": True},
        "datatables": {"class": "DataTable"}
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)
        # create instance methods for certain class_methods
        instance_methods = ["get"]
        instance_overload(self, instance_methods)

    def _get_url_instance(self, project_id=None):
        project_url = Sample.fm.get_project_url(project_id)
        url = f"{project_url}/samples/{self.id}"
        return url

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/samples/{id}"
        return url

    @classmethod
    def _get_all_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/samples"
        return url

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/samples"
        return url

    @classmethod
    def get(cls, id: str = "", name: str = "", shortid: str = "", project_id: str = "") -> Self:
        """
        Method to get one sample by its id, short_id, or name.
        If called on an instance of the class, the id of the class is used

        Parameters
        ----------
        id : str
            The id or short_id of the requested sample.
        name : str
            The name of the requested sample.
        shortid : str
            The shortid of the requested sample.
        project_id : str, optional
            Optionally to get the sample from another project as the furthrmind sdk was initiated with

        Returns
        -------
        Self
            An instance of the sample class.

        Raises
        ------
        AssertionError
            If called as a class method and no id, shortid, or name is specified.

        """

        if isclass(cls):
            assert id or name or shortid, "Either id, shortid, or name must be specified"

        return cls._get(id, shortid, name, project_id=project_id)

    @classmethod
    def get_many(cls, ids: List[str] = (), shortids: List[str] = (), names: List[str] = (),
                 project_id: str = None) -> List[Self]:
        """
        Method to get many samples by its ids, short_ids, or names.

        Parameters
        ----------
        ids : List[str]
            List of sample ids to filter samples by.
        shortids : List[str]
            List of short ids to filter samples by.
        names : List[str]
            List of names to filter samples by.
        project_id : str, optional
            Optionally to get samples from another project as the furthrmind sdk was initiated with

        Returns
        -------
        List[Self]
            List of instances of the sample class.

        Raises
        ------
        AssertionError
            If neither ids, shortids, nor names are specified.

        """

        assert ids or names or shortids, "Either ids, shortids, or names must be specified"
        return cls._get_many(ids, shortids, names, project_id=project_id)

    @classmethod
    def get_all(cls, project_id: str = "") -> List[Self]:
        """
        Method to get all samples belonging to one project

        Parameters
        ----------
        project_id : str, optional
            Optionally to get samples from another project as the furthrmind sdk was initiated with

        Returns
        -------
        List[Self]
            List of instances of the `Sample` class representing the fetched samples.
        """

        return cls._get_all(project_id)

    @classmethod
    @BaseClass._create_instances_decorator(_fetched=False)
    def create(cls, name, group_name=None, group_id=None, project_id=None) -> Self:
        """
        Method to create a new sample

        :param name: the name of the item to be created
        :param group_name: The name of the group where the new item will belong to. group name can be only considered
            for groups that are not subgroups. Either group_name or group_id must be specified
        :param group_id: the id of the group where the new item will belong to. Either group_name or group_id must be specified
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return instance of the sample class

        """

        return Sample._create(name, group_name, group_id, project_id)

    @classmethod
    @BaseClass._create_instances_decorator(_fetched=False)
    def create_many(cls, data_list: List[Dict], project_id=None) -> Self:
        """
        Method to create multiple samples

        :param data_list: dict with the following keys:
            - name: the name of the item to be created
            - group_name: The name of the group where the new item will belong to. group name can be only considered
            for groups that are not subgroups. Either group_name or group_id must be specified
            - group_id: the id of the group where the new item will belong to. Either group_name or group_id must be specified
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return list with instance of the sample class

        """

        return Sample._create_many(data_list, project_id)

    def add_datatable(self, name: str, columns: List[Dict] = None, project_id=None) -> "DataTable":
        """
        Method to create a new datatable within this sample

        :param name: name of the datatable
        :param columns: a list of columns that should be added to the datatable. List with dicts with the following keys:
            - name: name of the column
            - type: Type of the column, Either "Text" or "Numeric". Data must fit to type, for Text all data
            will be converted to string and for Numeric all data is converted to float (if possible)
            - data: List of column values, must fit to column_type
            - unit: dict with id or name, or name as string, or id as string
        :param project_id: Optionally to create an item in another project as the furthrmind sdk was initiated with
        :return: instance of column datatable class

        """

        from furthrmind.collection import DataTable
        datatable = DataTable.create(name, sample_id=self.id, columns=columns, project_id=project_id)

        new_datatable = list(self.datatables)
        new_datatable.append(datatable)
        self.datatables = new_datatable

        return datatable



from furthrmind.collection.baseclass import BaseClassWithFieldData, BaseClass
from typing_extensions import List, Dict, Self, TYPE_CHECKING
from inspect import isclass
if TYPE_CHECKING:
    from furthrmind.collection import *


class Group(BaseClassWithFieldData):
    id = ""
    name = ""
    neglect = False
    shortid = ""
    files: List["File"] = []
    fielddata: List["FieldData"] = []
    experiments: List["Experiment"] = []
    samples: List["Sample"] = []
    researchitems: Dict[str, List["ResearchItem"]] = {}
    sub_groups: List[Self] = []
    parent_group: Self = None

    _attr_definition = {
        "files": {"class": "File"},
        "fielddata": {"class": "FieldData"},
        "samples": {"class": "Sample"},
        "experiments": {"class": "Experiment"},
        "researchitems": {"class": "ResearchItem", "nested_dict": True},
        "sub_groups": {"class": "Group"},
        "parent_group": {"class": "Group"}
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    def _get_url_instance(self, project_id=None):
        project_url = Group.fm.get_project_url(project_id)
        url = f"{project_url}/groups/{self.id}"
        return url

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/groups/{id}"
        return url

    @classmethod
    def _get_all_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/groups"
        return url

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/groups"
        return url

    @classmethod
    def get(cls, id: str = "", name: str = "", shortid: str = "", parent_group_id: str = "", project_id: str = "") -> Self:
        """
        Method to get one group by its id, name, or shortid.
        If called on an instance of the class, the id of the class is used.

        Parameters
        ----------
        id : str
            id or short_id of requested group
        name : str
            name of requested group. For subgroups, see parent_group_id parameter
        shortid : str
            shortid of requested group
        parent_group_id : str
            id of parent group.
            If a subgroup is requested, the name and the parent_group_id is required
        project_id : str, optional
            Optionally to get a group from another project as the furthrmind sdk was initiated with.
            Defaults to None

        Returns
        -------
        Self
            Instance of group class

        Raises
        ------
        AssertionError
            When used as a class method, id, name, or shortid must be specified.

        """

        if isclass(cls):
            assert id or name or shortid, "Either id or name must be specified"

        return cls._get(id, shortid, name, parent_group_id=parent_group_id, project_id=project_id)

    @classmethod
    def get_many(cls, ids: List[str] = (), shortids: List[str] = (), names: List[str] = (),
                 project_id: str = "") -> List[Self]:
        """
        Method to get many groups

        Parameters
        ----------
        ids : List[str]
            List of ids to filter the groups.
        shortids : List[str]
            List of short_ids to filter the groups.
        names : List[str]
            List of names to filter the groups.
        project_id : str, optional
            Optionally to get groups from another project as the furthrmind sdk was initiated with

        Returns
        -------
        List[Self]
            List of instances of the Group class.

        Raises
        ------
        AssertionError
            If none of the parameters (ids, shortids, or names) are specified.

        """

        assert ids or names or shortids, "Either ids, shortids, or names must be specified"
        return cls._get_many(ids, shortids, names, project_id=project_id)

    @classmethod
    def get_all(cls, project_id: str = "") -> List[Self]:
        """
        Method to get all groups belonging to one project

        Parameters
        ----------
        project_id : str, optional
            Optionally to get groups from another project as the furthrmind sdk was initiated with

        Returns
        -------
        List[Self]
            A list of instances of the group class.

        """

        return super()._get_all(project_id)
    
    @classmethod
    @BaseClass._create_instances_decorator(_fetched=False)
    def create(cls, name: str, parent_group_id: str = "", project_id: str = "") -> Self:
        """
        Method to create a new group.

        Parameters
        ----------
        name: str
            The name of the item to be created.
        parent_group_id: str, optional
            The ID of the parent group where the new group should belong to. Defaults to an empty string.
        project_id: str, optional
            Optionally, create a group in another project as the furthrmind SDK was initiated with

        Returns
        -------
        instance of the group class
        """

        data = {"name": name}
        if parent_group_id:
            data["parent_group"] = {"id": parent_group_id}
        id = cls._post(data, project_id)
        data["id"] = id
        return data

    @classmethod
    @BaseClass._create_instances_decorator(_fetched=False)
    def create_many(cls, name_list: List[str], project_id: str = "") -> List[Self]:
        """
        Method to create multiple groups

        Parameters
        ----------
        name_list: List[str]
            A list containing names of the groups to be created.
        project_id: str, optional
            Optionally to create groups in another project as the furthrmind sdk was initiated with

        Returns
        -------
        List[Self]
            A list containing instances of the group class corresponding to the groups created.

        """

        data_list = [{"name": name} for name in name_list]

        id_list = cls._post(data_list, project_id, force_list=True)

        for data, id in zip(data_list, id_list):
            data["id"] = id

        return data_list



from inspect import isclass

from typing_extensions import Self, List, Dict, TYPE_CHECKING

from furthrmind.collection.baseclass import (BaseClassWithFieldData,
                                             BaseClassWithFiles,
                                             BaseClassWithGroup,
                                             BaseClassWithLinking,
                                             BaseClass
                                             )

if TYPE_CHECKING:
    from furthrmind.collection import FieldData, Sample, Group, ResearchItem, DataTable, File


class Experiment(BaseClassWithFieldData, BaseClassWithFiles,
                 BaseClassWithGroup, BaseClassWithLinking):
    id = ""
    name = ""
    neglect = False
    shortid = ""
    files: List["File"] = []
    fielddata: List["FieldData"] = []
    linked_samples: List["Sample"] = []
    linked_experiments: List[Self] = []
    groups: List["Group"] = []
    linked_researchitems: Dict[str, List["ResearchItem"]] = []
    datatables: List["DataTable"] = []

    _attr_definition = {
        "files": {"class": "File"},
        "fielddata": {"class": "FieldData"},
        "linked_samples": {"class": "Sample"},
        "linked_experiments": {"class": "Experiment"},
        "groups": {"class": "Group"},
        "linked_researchitems": {"class": "ResearchItem", "nested_dict": True},
        "datatables": {"class": "DataTable"}
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    def _get_url_instance(self, project_id=None):
        project_url = Experiment.fm.get_project_url(project_id)
        url = f"{project_url}/experiments/{self.id}"
        return url

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/experiments/{id}"
        return url

    @classmethod
    def _get_all_url(cls, project_id: str = None) -> str:
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/experiments"
        return url

    @classmethod
    def _post_url(cls, project_id=None):
        project_url = cls.fm.get_project_url(project_id)
        url = f"{project_url}/experiments"
        return url

    @classmethod
    def get(cls, id: str = None, name: str = None, shortid: str = "", project_id: str = "") -> Self:
        """
        Method to get one experiment by its id, name, or short_id
        If called on an instance of the class, the id of the class is used

        Parameters
        ----------
        id : str, optional
            The ID of the experiment to retrieve. If not provided, the ID of the calling instance will be used.
        name : str, optional
            The name of the experiment to retrieve.
        shortid : str, optional
            The short ID of the experiment to retrieve.
        project_id : str, optional
            Optionally to get an experiment from another project as the furthrmind sdk was initiated with

        Returns
        -------
        Self
            An instance of the experiment class.

        Raises
        ------
        AssertionError
            If neither ID nor name nor shortid is specified.

        Example usage:
            experiment = Experiment.get(id='ex123')
        """

        if isclass(cls):
            assert id or name or shortid, "Either id, name, or shortid must be specified"

        return cls._get(id, shortid, name, project_id=project_id)


    @classmethod
    def get_many(cls, ids: List[str] = (), shortids: List[str] = (), names: List[str] = (), project_id: str = "") -> List[Self]:
        """
        Method to get all experiment belonging to one project

        Parameters
        ----------
        ids : List[str]
            List of experiment ids.
        shortids : List[str]
            List of experiment short ids.
        names : List[str]
            List of experiment names.
        project_id : str, optional
            Optionally to get experiments from another project as the furthrmind sdk was initiated with

        Returns
        -------
        List[Self]
            List of instances of the experiment class.

        Raises
        ------
        AssertionError
            If none of the ids, shortids, or names are specified.

        """

        assert ids or names or shortids, "Either ids, shortids, or names must be specified"
        return cls._get_many(ids, shortids, names, project_id=project_id)

    @classmethod
    def get_all(cls, project_id: str = "") -> List[Self]:
        """
        Method to get all experiment belonging to one project

        Parameters
        ----------
        project_id : str, optional
            Optionally to get experiments from another project as the furthrmind sdk was initiated with

        Returns
        -------
        List[Self]
            A list containing instances of the experiment class.
        """

        return cls._get_all(project_id=project_id)

    @classmethod
    @BaseClass._create_instances_decorator(_fetched=False)
    def create(cls, name: str, group_name: str = "", group_id: str = "", project_id: str = "") -> Self:
        """
        Method to create a new experiment

        Parameters
        ----------
        name : str
            The name of the item to be created
        group_name : str, optional
            The name of the group where the new item will belong to. Group name can be only considered for groups that
            are not subgroups. Either group_name or group_id must be specified.
        group_id : str, optional
            The id of the group where the new item will belong to. Either group_name or group_id must be specified.
        project_id : str, optional
            Optionally, to create an item in another project as the furthrmind sdk was initiated with.

        Returns
        -------
        Experiment instance
            The instance of the experiment class created.

        """

        return cls._create(name, group_name, group_id, project_id)

    @classmethod
    @BaseClass._create_instances_decorator(_fetched=False)
    def create_many(cls, data_list: List[Dict], project_id: str = "") -> List[Self]:
        """
        Parameters
        ----------
        data_list : List[Dict]
            List of dictionaries containing information about the experiments to be created.
            Each dictionary should have the following keys:
            - name: str
                The name of the experiment.
            - group_name : str
                The name of the group where the experiment will belong to.
                Only applicable for groups that are not subgroups. Either group_name or group_id must be specified.
            - group_id : str
                The ID of the group where the experiment will belong to. Either group_name or group_id must be specified.

        project_id : str, optional
            Optionally to create an item in another project as the furthrmind sdk was initiated with

        Returns
        -------
        List[Self]
            List of instances of the experiment class.

        """

        return cls._create_many(data_list, project_id)

    def add_datatable(self, name: str, columns: List[Dict], project_id: str = "") -> "DataTable":
        """
        Method to create a new datatable within this experiment. Add the created datatable to the datatables attribute

        Parameters
        ----------
        name : str
            Name of the datatable.
        columns : List[Dict]
            A list of columns that should be added to the datatable. Each column is represented as a dictionary with the following keys:
            - name : str
                Name of the column.
            - type : str
                Type of the column. Either "Text" or "Numeric". Data must fit the specified type.
            - data : Union[List[Union[str, float]], pandas.Series]
                List of column values. Data must fit the specified type of the column.
                For Text columns, the items must be convertable to strings
                For Numeric columns, the items must be convertable to floats.
                Can be a list or a pandas.Series.
            - unit : dict or str
                Unit of the column. It can be represented as either a dictionary with 'id' or 'name', or a string
                representing the name or id of the unit.
        project_id : str, optional
            Optionally, specify the id of another project to create the datatable in.

        Returns
        -------
        DataTable
            An instance of the DataTable class representing the created datatable.

        """

        from furthrmind.collection import DataTable
        datatable = DataTable.create(name, experiment_id=self.id, columns=columns, project_id=project_id)

        new_datatables = list(self.datatables)
        new_datatables.append(datatable)
        self.datatables = new_datatables

        return datatable



from furthrmind.collection.baseclass import BaseClass
from typing_extensions import List, Dict, Self, TYPE_CHECKING
from inspect import isclass
if TYPE_CHECKING:
    from furthrmind.collection import *

class Project(BaseClass):
    id = ""
    name = ""
    info = ""
    shortid = ""
    samples: List["Sample"] = []
    experiments: List["Experiment"] = []
    groups: List["Group"] = []
    units: List["Unit"] = []
    researchitems: Dict[str, List["ResearchItem"]] = {}
    permissions = []
    fields: List["Field"] = []

    _attr_definition = {
        "samples": {"class": "Sample"},
        "experiments": {"class": "Experiment"},
        "groups": {"class": "Group"},
        "units": {"class": "Unit"},
        "researchitems": {"class": "ResearchItem", "nested_dict": True},
        "fields": {"class": "Field"}
    }

    def __init__(self, id=None, data=None):
        super().__init__(id, data)

    def _get_url_instance(self):
        project_url = self.fm.get_project_url(self.id)
        return project_url

    @classmethod
    def _get_url_class(cls, id, project_id=None):
        project_url = cls.fm.get_project_url(id)
        return project_url

    @classmethod
    def _get_all_url(cls, project_id=None):
        return f"{cls.fm.base_url}/projects"

    @classmethod
    def _post_url(cls):
        return f"{cls.fm.base_url}/projects"
    
    @classmethod
    def get(cls, id: str = "", name: str = "") -> Self:
        """
        This method is used to get one project by its id or name.
        If called on an instance of the class, the id of the class is used.
        Either id or name must be specified.

        Parameters
        ----------
        id : str, optional
            id or short_id of the requested project.
            Default value is an empty string.
        name : str, optional
            name of the requested project.
            Default value is an empty string.

        Returns
        -------
        Self
            Instance of the project class.

        """

        if isclass(cls):
            assert id or name, "Either id or name must be specified"

        return cls._get(id=id, name=name)


    @classmethod
    def get_many(cls, ids: List[str] = (), names: List[str] = ()) -> List[Self]:
        """
        Method to get many projects

        Parameters
        ----------
        ids : List[str]
            List of ids.

        names : List[str]
            List of names.

        Returns
        -------
        List[Self]
            List of instances of the class.

        Raises
        ------
        AssertionError
            If neither ids nor names are specified.
        """
        pass

        assert ids or names, "Either ids or names must be specified"
        return cls._get_many(ids, names)

    @classmethod
    def get_all(cls) -> List[Self]:
        """
        Method to get all projects

        Returns
        -------
        List[Self]
            List of instances of the class.

        """

        return super()._get_all()

    @classmethod
    @BaseClass._create_instances_decorator(_fetched=False)
    def create(cls, name: str) -> Self:
        """
        Method to create a new project

        Parameters
        ----------
        name : str
            Name of the new project

        Returns
        -------
        Self
            Instance of the project class

        Raises
        ------
        ValueError
            If name is empty or None

        """

        if not name:
            raise ValueError("Name is required")
        data = {"name": name}
        id = cls._post(data)
        data["id"] = id
        return data




